.source-edit {
    font-family: monospace;
    font-size: large;
}

.debug {
    background-color: alpha(pink, 0.25);
}

.gradient {
    background-image: linear-gradient(to bottom left, #1f2d8f, #159eff);
    color: white;
}
