/* marshal.h
 *
 * Copyright 2026 Eva M
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "bge.h"

G_BEGIN_DECLS

void
_marshal_DIRECT__ARGS_DIRECT (GClosure                *closure,
                              GValue                  *return_value,
                              guint                    n_param_values,
                              const GValue            *param_values,
                              gpointer invocation_hint G_GNUC_UNUSED,
                              gpointer                 marshal_data);

void
_marshal_BOXED__ARGS_DIRECT (GClosure                *closure,
                             GValue                  *return_value,
                             guint                    n_param_values,
                             const GValue            *param_values,
                             gpointer invocation_hint G_GNUC_UNUSED,
                             gpointer                 marshal_data);

void
_marshal_BOOL__ARGS_DIRECT (GClosure                *closure,
                            GValue                  *return_value,
                            guint                    n_param_values,
                            const GValue            *param_values,
                            gpointer invocation_hint G_GNUC_UNUSED,
                            gpointer                 marshal_data);

void
_marshal_DOUBLE__ARGS_DIRECT (GClosure                *closure,
                              GValue                  *return_value,
                              guint                    n_param_values,
                              const GValue            *param_values,
                              gpointer invocation_hint G_GNUC_UNUSED,
                              gpointer                 marshal_data);

G_END_DECLS
